#include "ImgView.h"

ImgView::ImgView(int x, int y, int w, int h, const char *label) : Fl_Double_Window(x,y,w,h,label)
{
	imgBuf = NULL;
	origImg = NULL;
	curImg = NULL;
	
	imgLeft = 0;
	imgTop = 0;

	fltDesignUI = new FltDesignUI;
	fltDesignUI->imgView = this;

	int i;
	for (i=0;i<FLT_HEIGHT*FLT_WIDTH;i++)
	{
		fltKernel[i] = 0;
	}
	fltKernel[(FLT_HEIGHT/2)*FLT_WIDTH+(FLT_HEIGHT/2)]=1;
	scale = 1;
	offset = 0;

	InitFltDesignUI();

	fltDesignUI->image->value(1);
	fltDesignUI->selection->value(0);
	
	brushSelection = NULL;
	brushType = ROUND_BRUSH;
	brushSize = 10;
	brushRadius2 = brushSize*brushSize;

	brushOpacity = 1.0;	

	brushConfigUI = new BrushConfigUI;
	brushConfigUI->imgView = this;

	brushConfigUI->round->value(1);
	brushConfigUI->square->value(0);
	brushConfigUI->brushSize->value(brushSize);
	brushConfigUI->brushOpacity->value(brushOpacity);

	helpPageUI = new HelpPageUI();
	char helpInfo[2048];
	my_img_filter(helpInfo);
	helpPageUI->helpText->value(helpInfo);
}

void ImgView::FreeBuffer(void)
{
	if (imgBuf)
	{
		delete[] imgBuf;
		delete[] origImg;
		delete[] curImg;
		delete[] brushSelection;

		imgBuf = NULL;
		origImg = NULL;
		curImg = NULL;

		brushSelection = NULL;
	}
}

void ImgView::HideAll(void)
{
	fltDesignUI->hide();
	brushConfigUI->hide();
}

ImgView::~ImgView()
{
	FreeBuffer();

	delete fltDesignUI;
	delete brushConfigUI;
	delete helpPageUI;
}

void ImgView::OpenImage(void)
{
	char *filename = fl_file_chooser("choose a bitmap file", "*.tga", 0);

	if (filename)
	{
		OpenImage(filename);
	}
}

void ImgView::OpenImage(const char *filename)
{
	int i,j,total;

	FreeBuffer();	

	CByteImage rick;
	ReadFile(rick,filename);

	imgWidth = rick.Shape().width;
	imgHeight = rick.Shape().height;

	total = imgWidth*imgHeight;

	brushSelection = new unsigned char[total];
	memset(brushSelection,0,total*sizeof(unsigned char));

	total *= 3;

	imgBuf = new unsigned char [total];

	//Rick's .tga file format is in G,B,R order from bottom to top and from left to right;
	//our buffer is in R, G, B order from top to bottom and from left to right;
	for (j=0;j<imgHeight;j++)
	{
		for (i=0;i<imgWidth;i++)
		{
			int index = 3*(j*imgWidth+i);
			int x = i;
			int y = imgHeight-1-j;
			imgBuf[index+0] = rick.Pixel(x,y,2);
			imgBuf[index+1] = rick.Pixel(x,y,1);
			imgBuf[index+2] = rick.Pixel(x,y,0);
		}
	}

	origImg = new double [total];
	curImg = new double [total];
	
	for (i=0;i<total;i++)
	{
		origImg[i] = imgBuf[i];
	}
	memcpy(curImg,origImg,total*sizeof(double));


	if (imgWidth>w())
	{
		imgLeft = 0;
	}
	else
	{
		imgLeft = (w()-imgWidth)/2;
	}

	if (imgHeight>h())
	{
		imgTop = 0;
	}
	else
	{
		imgTop = (h()-imgHeight)/2;
	}

	redraw();
}

void ImgView::SaveImage(void)
{
	char *filename = fl_file_chooser("choose a bitmap file", "*.tga", 0);

	if (filename)
	{
		SaveImage(filename);
	}
}

void ImgView::SaveImage(const char *filename)
{
	int i,j;

	CByteImage rick(imgWidth, imgHeight, 3);	

	//our buffer is in R, G, B order from top to bottom and from left to right;
	//Rick's .tga file format is in G,B,R order from bottom to top and from left to right;
	for (j=0;j<imgHeight;j++)
	{
		for (i=0;i<imgWidth;i++)
		{
			int index = 3*(j*imgWidth+i);
			int x = i;
			int y = imgHeight-1-j;
			rick.Pixel(x,y,0) = imgBuf[index+2];
			rick.Pixel(x,y,1) = imgBuf[index+1];
			rick.Pixel(x,y,2) = imgBuf[index+0];
		}
	}

	WriteFile(rick,filename);
}

void ImgView::TryFilter(void)
{
	fltDesignUI->show();
	//cout<<"try filter!"<<endl;
}

void ImgView::LoadFilter(void)
{
	char *filename = fl_file_chooser("choose a filter file", "*.txt", 0);

	if (filename)
	{
		LoadFilter(filename);
	}
}

void ImgView::LoadFilter(const char *filename)
{
	FILE *fp = fopen(filename,"r");

	int i;
	for (i=0;i<FLT_WIDTH*FLT_HEIGHT;i++)
	{
		fscanf(fp,"%lf",fltKernel+i);
	}

	fscanf(fp,"%lf",&scale);
	fscanf(fp,"%lf",&offset);

	fclose(fp);

	InitFltDesignUI();
}

void ImgView::SaveFilter(void)
{
	char *filename = fl_file_chooser("choose a filter file", "*.txt", 0);

	if (filename)
	{
		SaveFilter(filename);
	}
}

void ImgView::SaveFilter(const char *filename)
{
	FILE *fp = fopen(filename,"w");

	int i,j;
	for (j=0;j<FLT_HEIGHT;j++)
	{
		for (i=0;i<FLT_WIDTH;i++)
		{
			fprintf(fp,"%lf ",fltKernel[j*FLT_WIDTH+i]);
		}
		fprintf(fp,"\n");
	}

	fprintf(fp,"%lf\n",scale);
	fprintf(fp,"%lf",offset);

	fclose(fp);	
}

void ImgView::PreviewFilter(void)
{
	if (imgBuf)
	{
		Filter();

		UpdateBuffer();
	}

	//cout<<"preview"<<endl;	
}

void ImgView::CancelFilter(void)
{
	if (imgBuf)
	{
		memcpy(curImg,origImg,imgWidth*imgHeight*3*sizeof(double));

		UpdateBuffer();
	}

	//cout<<"cancel"<<endl;	
}

void ImgView::AcceptFilter(void)
{
	if (imgBuf)
	{
		memcpy(origImg,curImg,imgWidth*imgHeight*3*sizeof(double));
	}

	//cout<<"accept"<<endl;
}

void ImgView::StopFilter(void)
{
	if (imgBuf)
	{
		memcpy(curImg,origImg,imgWidth*imgHeight*3*sizeof(double));

		UpdateBuffer();
	}

	fltDesignUI->hide();
	//cout<<"stop filter!"<<endl;
}

void ImgView::InitFltDesignUI(void)
{
	fltDesignUI->ele0->value(fltKernel[0]);
	fltDesignUI->ele1->value(fltKernel[1]);
	fltDesignUI->ele2->value(fltKernel[2]);
	fltDesignUI->ele3->value(fltKernel[3]);
	fltDesignUI->ele4->value(fltKernel[4]);
	fltDesignUI->ele5->value(fltKernel[5]);
	fltDesignUI->ele6->value(fltKernel[6]);
	fltDesignUI->ele7->value(fltKernel[7]);
	fltDesignUI->ele8->value(fltKernel[8]);
	fltDesignUI->ele9->value(fltKernel[9]);
	fltDesignUI->ele10->value(fltKernel[10]);
	fltDesignUI->ele11->value(fltKernel[11]);
	fltDesignUI->ele12->value(fltKernel[12]);
	fltDesignUI->ele13->value(fltKernel[13]);
	fltDesignUI->ele14->value(fltKernel[14]);
	fltDesignUI->ele15->value(fltKernel[15]);
	fltDesignUI->ele16->value(fltKernel[16]);
	fltDesignUI->ele17->value(fltKernel[17]);
	fltDesignUI->ele18->value(fltKernel[18]);
	fltDesignUI->ele19->value(fltKernel[19]);
	fltDesignUI->ele20->value(fltKernel[20]);
	fltDesignUI->ele21->value(fltKernel[21]);
	fltDesignUI->ele22->value(fltKernel[22]);
	fltDesignUI->ele23->value(fltKernel[23]);
	fltDesignUI->ele24->value(fltKernel[24]);

	fltDesignUI->scale->value(scale);
	fltDesignUI->offset->value(offset);
}

void ImgView::UpdateFilter()
{
	fltKernel[0] = fltDesignUI->ele0->value();
	fltKernel[1] = fltDesignUI->ele1->value();
	fltKernel[2] = fltDesignUI->ele2->value();
	fltKernel[3] = fltDesignUI->ele3->value();
	fltKernel[4] = fltDesignUI->ele4->value();
	fltKernel[5] = fltDesignUI->ele5->value();
	fltKernel[6] = fltDesignUI->ele6->value();
	fltKernel[7] = fltDesignUI->ele7->value();
	fltKernel[8] = fltDesignUI->ele8->value();
	fltKernel[9] = fltDesignUI->ele9->value();
	fltKernel[10] = fltDesignUI->ele10->value();
	fltKernel[11] = fltDesignUI->ele11->value();
	fltKernel[12] = fltDesignUI->ele12->value();
	fltKernel[13] = fltDesignUI->ele13->value();
	fltKernel[14] = fltDesignUI->ele14->value();
	fltKernel[15] = fltDesignUI->ele15->value();
	fltKernel[16] = fltDesignUI->ele16->value();
	fltKernel[17] = fltDesignUI->ele17->value();
	fltKernel[18] = fltDesignUI->ele18->value();
	fltKernel[19] = fltDesignUI->ele19->value();
	fltKernel[20] = fltDesignUI->ele20->value();
	fltKernel[21] = fltDesignUI->ele21->value();
	fltKernel[22] = fltDesignUI->ele22->value();
	fltKernel[23] = fltDesignUI->ele23->value();
	fltKernel[24] = fltDesignUI->ele24->value();

	scale = fltDesignUI->scale->value();
	offset = fltDesignUI->offset->value();

//	printf("update filter\n");
}

void ImgView::UpdateBuffer()
{
	if (imgBuf)
	{
		int i;
		int total = imgWidth*imgHeight;
		for (i=0;i<total;i++)
		{
			int index = 3*i;
			double tmp;

			if (brushSelection[i])
			{
				for (int c = 0; c<3; c++)
				{
					tmp = curImg[index+c];
					tmp = __min(255.0,tmp);
					tmp = __max(0.0,tmp);
					imgBuf[index+c] = (unsigned char)(int)(floor(tmp)*brushOpacity);
				}
			}
			else
			{
				for (int c = 0; c<3; c++)
				{
					tmp = curImg[index+c];
					tmp = __min(255.0,tmp);
					tmp = __max(0.0,tmp);
					imgBuf[index+c] = (unsigned char)(int)floor(tmp);
				}
			}
		}
		redraw();
	}
}

void ImgView::Filter(void)
{
	if (imgBuf)
	{
		unsigned char *range = NULL;

		if (fltDesignUI->selection->value())
		{
			range = brushSelection;
		}

		image_filter(curImg, origImg, range, imgWidth, imgHeight, 
			fltKernel, FLT_WIDTH, FLT_HEIGHT, 
			scale, offset);
	}
}

void ImgView::TryBrush(void)
{
	brushConfigUI->show();
}

void ImgView::UpdateBrushConfig(void)
{
	if (brushConfigUI->round->value())
	{
		brushType = ROUND_BRUSH;
	}
	else if (brushConfigUI->square->value())
	{
		brushType = SQUARE_BRUSH;
	}

	brushSize = (int)brushConfigUI->brushSize->value();
	brushOpacity = brushConfigUI->brushOpacity->value();

	brushRadius2 = brushSize * brushSize;

	//printf("brush size %d\n",brushSize);

	UpdateBuffer();
}

void ImgView::CleanBrushSelection(void)
{
	if (imgBuf)
	{
		memset(brushSelection,0,imgWidth*imgHeight*sizeof(unsigned char));
		UpdateBuffer();
	}
}

int ImgView::handle(int c)
{	
	if (c==FL_PUSH)
	{
		return 1;
	}
	else if (c==FL_MOVE)
	{
		int x = Fl::event_x();
		int y = Fl::event_y();

		int col = x-imgLeft;
		int row = y-imgTop;

		unsigned char pixel[3] = {BK_COLOR[0],BK_COLOR[1],BK_COLOR[2]};
		if (imgBuf
			&&0<=col&&col<imgWidth
			&&0<=row&&row<imgHeight)
		{
			memcpy(pixel,imgBuf+3*(row*imgWidth+col),3*sizeof(unsigned char));
		}

		char info[256];
		sprintf(info,"x = %d y = %d r = %d g = %d b = %d",
			col,row,(int)pixel[0],(int)pixel[1],(int)pixel[2]);
		mouseInfo->value(info);

		mouseX = x;
		mouseY = y;

		return 1;
	}	
	else if (c==FL_DRAG)
	{
		int x = Fl::event_x();
		int y = Fl::event_y();

		if (Fl::event_button()==FL_RIGHT_MOUSE)
		{
			imgLeft += x - mouseX;
			imgTop += y - mouseY;
			redraw();
		}
		else if (Fl::event_button()==FL_LEFT_MOUSE)
		{
			int cntX = x - imgLeft;
			int cntY = y - imgTop;

			/************************ TO BE DONE 3 **************************/
			//cntX and cntY are mouse position in image coordinates:
			//you are required to first filter the image in the range of
			//[ cntX-brushSize , cntX+brushSize ] * [ cntY-brushSize , cntY+brushSize ]
			//Then update the imgBuf in this range accordingly

			printf("brush: to be done in function handle in file ImgView.cpp\n");

			/************************ END OF TBD 3 **************************/
			
			redraw();
		}

		mouseX = x;
		mouseY = y;

		return 1;
	}
	else if (c==FL_RELEASE)
	{
		return 1;
	}
	else
	{
		return Fl_Window::handle(c);
	}
}

void ImgView::draw(void)
{
	fl_rectf(0,0,w(),h(),BK_COLOR[0],BK_COLOR[1],BK_COLOR[2]);

	if (imgBuf)
	{
		fl_draw_image(imgBuf, imgLeft, imgTop, imgWidth, imgHeight, 3);
	}
	else
	{
		Fl_Window::draw();
	}
}

void ImgView::AboutMe(void)
{
	helpPageUI->show();
}
