#ifndef IMGVIEW_H
#define IMGVIEW_H

#include "imgflt.h"

const unsigned char BK_COLOR[3] = {212,208,200};
const int FLT_WIDTH = 5;
const int FLT_HEIGHT = 5;

const int ROUND_BRUSH = 1;
const int SQUARE_BRUSH = 2;

class ImgView : public Fl_Double_Window
{
protected:	

	//Three copies of are kept for an image:
	//imgBuf:				a byte buffer used for rendering
	//origImg:				a real buffer of original image, used for floating point operation
	//curImg:				a real buffer used to save the result of processing orgImg, 
	//						if a certain procedure, e.g., filtering, is accepted, curImg is copied to origImg.
	//imgWidth, imgHeight:	dimension of the image.
	unsigned char *imgBuf;
	double *origImg, *curImg;
	int imgWidth, imgHeight;

	int imgLeft,imgTop;		//the position of the topleft pixel in the window
	int mouseX,mouseY;		//the mouse position in the current in the window

	double fltKernel[FLT_WIDTH*FLT_HEIGHT];		//the kernel of the image filter
	double scale,offset;						//the scale and offset parameter, 
												//applied after convolving to transform 
												//the filtered image into the range [0,255]
	
	unsigned char *brushSelection;				//a binary array, indicating which area is selected by brush
	int brushType;								//brushType is either ROUND or SQUARE
	int brushSize;								//for a round brush, brushSize is the radius of the cicle
												//for a square brush, brushSize is half of the side length
	int brushRadius2;							//brushRadius2 = brushSize * brushSize
												//the default brush is a round brush of size 10

	double brushOpacity;						//brush selected area can be rendered with opacity, ranging in [0,1]
												//the default opacity if 1.0, which is transparent. 
												//the opacity can be adusted in brush config pannel.

public:

	// Those are four UI windows for infomation prompting, and parameter adjusting;
	Fl_Output *mouseInfo;
	FltDesignUI *fltDesignUI;	
	BrushConfigUI *brushConfigUI;
	HelpPageUI *helpPageUI;						

public:
	ImgView(int x, int y, int w, int h, const char *label=0);
	virtual ~ImgView();

	//the following two functions are called when menuitems in File menu are selected
	void OpenImage(void);	
	void SaveImage(void);	

	//TryBrush is called when Tools-->Filter menu is selected
	void TryFilter(void);
	//the following six functions are called when respective buttons are 
	//pressed in the filter design panel.
	void LoadFilter(void);
	void SaveFilter(void);	
	void PreviewFilter(void);
	void CancelFilter(void);
	void AcceptFilter(void);
	void StopFilter(void);
	//UpdateFilter is called whenever filter property is modified through the filter design panel
	void UpdateFilter(void);

	//TryBrush is called when Tools-->Brush menu is selected
	void TryBrush(void);	
	//UpdateBrushConfig is called whenever brush property is modified through
	//the brush config panel
	void UpdateBrushConfig(void);
	//CleanBrushSelection is called when the "clean" button is pressed in the 
	//the brush config panel, to clean all the bookmarking for previous selected brush area.
	void CleanBrushSelection(void);
	
	//two utility routines:
	int IsPtInImage(int i, int j) const 
	{ return 0<=i && i<imgWidth && 0<=j && j<imgHeight; }
	int IsPtInRoundBrush(int i, int j) const 
	{ return i*i+j*j<brushRadius2; }

	//HideAll is called before exit, closing filter design panel and brush config panel
	void HideAll(void);

	//draw displays proper image as desired.
	void draw(void);

/**************************************************************************************/
	//handle processes all the mouse, keyboard messages to support user interface.
	//TO SUPPORT BRUSH FEATURE, YOU NEED TO WORK IN THIS FUNCTION!!!!!
/**************************************************************************************/
	int handle(int c);

	//called if help menu item is selected.
	void AboutMe();

protected:

	//the following functions help to implement file I/O, memory allocate/copy/free
	//routines required in the public interface function;

	void OpenImage(const char *filename);
	void SaveImage(const char *filename);

	void LoadFilter(const char *filename);
	void InitFltDesignUI(void);
	void SaveFilter(const char *filename);	
	
	void UpdateBuffer(void);
	void FreeBuffer(void);

	//the Filter calls the filter function you write to actually *filter* images.
	void Filter(void);
};

#endif